@@ -29,7 +29,10 @@ import itertools
 import math
 import os
 import subprocess
+import sys
 from enum import IntEnum
+from functools import cached_property
+from typing import IO, TYPE_CHECKING, Any, List, Literal, NamedTuple, Union
 
 from . import (
     Image,
@@ -44,6 +47,9 @@ from ._binary import i16le as i16
 from ._binary import o8
 from ._binary import o16le as o16
 
+if TYPE_CHECKING:
+    from . import _imaging
+
 
 class LoadingStrategy(IntEnum):
     """.. versionadded:: 9.1.0"""
@@ -60,7 +66,7 @@ LOADING_STRATEGY = LoadingStrategy.RGB_AFTER_FIRST
 # Identify/read GIF files
 
 
-def _accept(prefix):
+def _accept(prefix: bytes) -> bool:
     return prefix[:6] in [b"GIF87a", b"GIF89a"]
 
 
@@ -76,19 +82,19 @@ class GifImageFile(ImageFile.ImageFile):
 
     global_palette = None
 
-    def data(self):
+    def data(self) -> bytes | None:
         s = self.fp.read(1)
         if s and s[0]:
             return self.fp.read(s[0])
         return None
 
-    def _is_palette_needed(self, p):
+    def _is_palette_needed(self, p: bytes) -> bool:
         for i in range(0, len(p), 3):
             if not (i // 3 == p[i] == p[i + 1] == p[i + 2]):
                 return True
         return False
 
-    def _open(self):
+    def _open(self) -> None:
         # Screen
         s = self.fp.read(13)
         if not _accept(s):
@@ -112,12 +118,11 @@ class GifImageFile(ImageFile.ImageFile):
 
         self._fp = self.fp  # FIXME: hack
         self.__rewind = self.fp.tell()
-        self._n_frames = None
-        self._is_animated = None
+        self._n_frames: int | None = None
         self._seek(0)  # get ready to read first frame
 
     @property
-    def n_frames(self):
+    def n_frames(self) -> int:
         if self._n_frames is None:
             current = self.tell()
             try:
@@ -128,26 +133,25 @@ class GifImageFile(ImageFile.ImageFile):
             self.seek(current)
         return self._n_frames
 
-    @property
-    def is_animated(self):
-        if self._is_animated is None:
-            if self._n_frames is not None:
-                self._is_animated = self._n_frames != 1
-            else:
-                current = self.tell()
-                if current:
-                    self._is_animated = True
-                else:
-                    try:
-                        self._seek(1, False)
-                        self._is_animated = True
-                    except EOFError:
-                        self._is_animated = False
+    @cached_property
+    def is_animated(self) -> bool:
+        if self._n_frames is not None:
+            return self._n_frames != 1
 
-                    self.seek(current)
-        return self._is_animated
+        current = self.tell()
+        if current:
+            return True
 
-    def seek(self, frame):
+        try:
+            self._seek(1, False)
+            is_animated = True
+        except EOFError:
+            is_animated = False
+
+        self.seek(current)
+        return is_animated
+
+    def seek(self, frame: int) -> None:
         if not self._seek_check(frame):
             return
         if frame < self.__frame:
@@ -163,11 +167,11 @@ class GifImageFile(ImageFile.ImageFile):
                 msg = "no more images in GIF file"
                 raise EOFError(msg) from e
 
-    def _seek(self, frame, update_image=True):
+    def _seek(self, frame: int, update_image: bool = True) -> None:
         if frame == 0:
             # rewind
             self.__offset = 0
-            self.dispose = None
+            self.dispose: _imaging.ImagingCore | None = None
             self.__frame = -1
             self._fp.seek(self.__rewind)
             self.disposal_method = 0
@@ -195,9 +199,9 @@ class GifImageFile(ImageFile.ImageFile):
             msg = "no more images in GIF file"
             raise EOFError(msg)
 
-        palette = None
+        palette: ImagePalette.ImagePalette | Literal[False] | None = None
 
-        info = {}
+        info: dict[str, Any] = {}
         frame_transparency = None
         interlace = None
         frame_dispose_extent = None
@@ -213,7 +217,7 @@ class GifImageFile(ImageFile.ImageFile):
                 #
                 s = self.fp.read(1)
                 block = self.data()
-                if s[0] == 249:
+                if s[0] == 249 and block is not None:
                     #
                     # graphic control extension
                     #
@@ -249,14 +253,14 @@ class GifImageFile(ImageFile.ImageFile):
                         info["comment"] = comment
                     s = None
                     continue
-                elif s[0] == 255 and frame == 0:
+                elif s[0] == 255 and frame == 0 and block is not None:
                     #
                     # application extension
                     #
                     info["extension"] = block, self.fp.tell()
                     if block[:11] == b"NETSCAPE2.0":
                         block = self.data()
-                        if len(block) >= 3 and block[0] == 1:
+                        if block and len(block) >= 3 and block[0] == 1:
                             self.info["loop"] = i16(block, 1)
                 while self.data():
                     pass
@@ -337,60 +341,60 @@ class GifImageFile(ImageFile.ImageFile):
                         self._mode = "RGB"
                         self.im = self.im.convert("RGB", Image.Dither.FLOYDSTEINBERG)
 
-        def _rgb(color):
+        def _rgb(color: int) -> tuple[int, int, int]:
             if self._frame_palette:
                 if color * 3 + 3 > len(self._frame_palette.palette):
                     color = 0
-                color = tuple(self._frame_palette.palette[color * 3 : color * 3 + 3])
+                return tuple(self._frame_palette.palette[color * 3 : color * 3 + 3])
             else:
-                color = (color, color, color)
-            return color
+                return (color, color, color)
 
+        self.dispose = None
         self.dispose_extent = frame_dispose_extent
-        try:
-            if self.disposal_method < 2:
-                # do not dispose or none specified
-                self.dispose = None
-            elif self.disposal_method == 2:
-                # replace with background colour
+        if self.dispose_extent and self.disposal_method >= 2:
+            try:
+                if self.disposal_method == 2:
+                    # replace with background colour
 
-                # only dispose the extent in this frame
-                x0, y0, x1, y1 = self.dispose_extent
-                dispose_size = (x1 - x0, y1 - y0)
-
-                Image._decompression_bomb_check(dispose_size)
-
-                # by convention, attempt to use transparency first
-                dispose_mode = "P"
-                color = self.info.get("transparency", frame_transparency)
-                if color is not None:
-                    if self.mode in ("RGB", "RGBA"):
-                        dispose_mode = "RGBA"
-                        color = _rgb(color) + (0,)
-                else:
-                    color = self.info.get("background", 0)
-                    if self.mode in ("RGB", "RGBA"):
-                        dispose_mode = "RGB"
-                        color = _rgb(color)
-                self.dispose = Image.core.fill(dispose_mode, dispose_size, color)
-            else:
-                # replace with previous contents
-                if self.im is not None:
                     # only dispose the extent in this frame
-                    self.dispose = self._crop(self.im, self.dispose_extent)
-                elif frame_transparency is not None:
                     x0, y0, x1, y1 = self.dispose_extent
                     dispose_size = (x1 - x0, y1 - y0)
 
                     Image._decompression_bomb_check(dispose_size)
+
+                    # by convention, attempt to use transparency first
                     dispose_mode = "P"
-                    color = frame_transparency
-                    if self.mode in ("RGB", "RGBA"):
-                        dispose_mode = "RGBA"
-                        color = _rgb(frame_transparency) + (0,)
+                    color = self.info.get("transparency", frame_transparency)
+                    if color is not None:
+                        if self.mode in ("RGB", "RGBA"):
+                            dispose_mode = "RGBA"
+                            color = _rgb(color) + (0,)
+                    else:
+                        color = self.info.get("background", 0)
+                        if self.mode in ("RGB", "RGBA"):
+                            dispose_mode = "RGB"
+                            color = _rgb(color)
                     self.dispose = Image.core.fill(dispose_mode, dispose_size, color)
-        except AttributeError:
-            pass
+                else:
+                    # replace with previous contents
+                    if self.im is not None:
+                        # only dispose the extent in this frame
+                        self.dispose = self._crop(self.im, self.dispose_extent)
+                    elif frame_transparency is not None:
+                        x0, y0, x1, y1 = self.dispose_extent
+                        dispose_size = (x1 - x0, y1 - y0)
+
+                        Image._decompression_bomb_check(dispose_size)
+                        dispose_mode = "P"
+                        color = frame_transparency
+                        if self.mode in ("RGB", "RGBA"):
+                            dispose_mode = "RGBA"
+                            color = _rgb(frame_transparency) + (0,)
+                        self.dispose = Image.core.fill(
+                            dispose_mode, dispose_size, color
+                        )
+            except AttributeError:
+                pass
 
         if interlace is not None:
             transparency = -1
@@ -417,7 +421,7 @@ class GifImageFile(ImageFile.ImageFile):
             elif k in self.info:
                 del self.info[k]
 
-    def load_prepare(self):
+    def load_prepare(self) -> None:
         temp_mode = "P" if self._frame_palette else "L"
         self._prev_im = None
         if self.__frame == 0:
@@ -429,7 +433,7 @@ class GifImageFile(ImageFile.ImageFile):
             self._prev_im = self.im
             if self._frame_palette:
                 self.im = Image.core.fill("P", self.size, self._frame_transparency or 0)
-                self.im.putpalette(*self._frame_palette.getdata())
+                self.im.putpalette("RGB", *self._frame_palette.getdata())
             else:
                 self.im = None
         self._mode = temp_mode
@@ -437,7 +441,7 @@ class GifImageFile(ImageFile.ImageFile):
 
         super().load_prepare()
 
-    def load_end(self):
+    def load_end(self) -> None:
         if self.__frame == 0:
             if self.mode == "P" and LOADING_STRATEGY == LoadingStrategy.RGB_ALWAYS:
                 if self._frame_transparency is not None:
@@ -454,6 +458,8 @@ class GifImageFile(ImageFile.ImageFile):
             frame_im = self.im.convert("RGBA")
         else:
             frame_im = self.im.convert("RGB")
+
+        assert self.dispose_extent is not None
         frame_im = self._crop(frame_im, self.dispose_extent)
 
         self.im = self._prev_im
@@ -463,7 +469,7 @@ class GifImageFile(ImageFile.ImageFile):
         else:
             self.im.paste(frame_im, self.dispose_extent)
 
-    def tell(self):
+    def tell(self) -> int:
         return self.__frame
 
 
@@ -474,7 +480,7 @@ class GifImageFile(ImageFile.ImageFile):
 RAWMODE = {"1": "L", "L": "L", "P": "P"}
 
 
-def _normalize_mode(im):
+def _normalize_mode(im: Image.Image) -> Image.Image:
     """
     Takes an image (or frame), returns an image in a mode that is appropriate
     for saving in a Gif.
@@ -499,7 +505,12 @@ def _normalize_mode(im):
     return im.convert("L")
 
 
-def _normalize_palette(im, palette, info):
+_Palette = Union[bytes, bytearray, List[int], ImagePalette.ImagePalette]
+
+
+def _normalize_palette(
+    im: Image.Image, palette: _Palette | None, info: dict[str, Any]
+) -> Image.Image:
     """
     Normalizes the palette for image.
       - Sets the palette to the incoming palette, if provided.
@@ -527,8 +538,10 @@ def _normalize_palette(im, palette, info):
             source_palette = bytearray(i // 3 for i in range(768))
         im.palette = ImagePalette.ImagePalette("RGB", palette=source_palette)
 
+    used_palette_colors: list[int] | None
     if palette:
         used_palette_colors = []
+        assert source_palette is not None
         for i in range(0, len(source_palette), 3):
             source_color = tuple(source_palette[i : i + 3])
             index = im.palette.colors.get(source_color)
@@ -559,7 +572,11 @@ def _normalize_palette(im, palette, info):
     return im
 
 
-def _write_single_frame(im, fp, palette):
+def _write_single_frame(
+    im: Image.Image,
+    fp: IO[bytes],
+    palette: _Palette | None,
+) -> None:
     im_out = _normalize_mode(im)
     for k, v in im_out.info.items():
         im.encoderinfo.setdefault(k, v)
@@ -580,7 +597,9 @@ def _write_single_frame(im, fp, palette):
     fp.write(b"\0")  # end of image data
 
 
-def _getbbox(base_im, im_frame):
+def _getbbox(
+    base_im: Image.Image, im_frame: Image.Image
+) -> tuple[Image.Image, tuple[int, int, int, int] | None]:
     if _get_palette_bytes(im_frame) != _get_palette_bytes(base_im):
         im_frame = im_frame.convert("RGBA")
         base_im = base_im.convert("RGBA")
@@ -588,12 +607,20 @@ def _getbbox(base_im, im_frame):
     return delta, delta.getbbox(alpha_only=False)
 
 
-def _write_multiple_frames(im, fp, palette):
+class _Frame(NamedTuple):
+    im: Image.Image
+    bbox: tuple[int, int, int, int] | None
+    encoderinfo: dict[str, Any]
+
+
+def _write_multiple_frames(
+    im: Image.Image, fp: IO[bytes], palette: _Palette | None
+) -> bool:
     duration = im.encoderinfo.get("duration")
     disposal = im.encoderinfo.get("disposal", im.info.get("disposal"))
 
-    im_frames = []
-    previous_im = None
+    im_frames: list[_Frame] = []
+    previous_im: Image.Image | None = None
     frame_count = 0
     background_im = None
     for imSequence in itertools.chain([im], im.encoderinfo.get("append_images", [])):
@@ -619,43 +646,48 @@ def _write_multiple_frames(im, fp, palette):
             frame_count += 1
 
             diff_frame = None
-            if im_frames:
+            if im_frames and previous_im:
                 # delta frame
                 delta, bbox = _getbbox(previous_im, im_frame)
                 if not bbox:
                     # This frame is identical to the previous frame
                     if encoderinfo.get("duration"):
-                        im_frames[-1]["encoderinfo"]["duration"] += encoderinfo[
-                            "duration"
-                        ]
+                        im_frames[-1].encoderinfo["duration"] += encoderinfo["duration"]
                     continue
-                if encoderinfo.get("disposal") == 2:
+                if im_frames[-1].encoderinfo.get("disposal") == 2:
                     if background_im is None:
                         color = im.encoderinfo.get(
                             "transparency", im.info.get("transparency", (0, 0, 0))
                         )
                         background = _get_background(im_frame, color)
                         background_im = Image.new("P", im_frame.size, background)
-                        background_im.putpalette(im_frames[0]["im"].palette)
-                    delta, bbox = _getbbox(background_im, im_frame)
-                if encoderinfo.get("optimize") and im_frame.mode != "1":
+                        background_im.putpalette(im_frames[0].im.palette)
+                    bbox = _getbbox(background_im, im_frame)[1]
+                elif encoderinfo.get("optimize") and im_frame.mode != "1":
                     if "transparency" not in encoderinfo:
                         try:
-                            encoderinfo[
-                                "transparency"
-                            ] = im_frame.palette._new_color_index(im_frame)
+                            encoderinfo["transparency"] = (
+                                im_frame.palette._new_color_index(im_frame)
+                            )
                         except ValueError:
                             pass
                     if "transparency" in encoderinfo:
                         # When the delta is zero, fill the image with transparency
                         diff_frame = im_frame.copy()
-                        fill = Image.new(
-                            "P", diff_frame.size, encoderinfo["transparency"]
-                        )
+                        fill = Image.new("P", delta.size, encoderinfo["transparency"])
                         if delta.mode == "RGBA":
                             r, g, b, a = delta.split()
-                            mask = ImageMath.eval(
-                                "convert(max(max(max(r, g), b), a) * 255, '1')",
+                            mask = ImageMath.lambda_eval(
+                                lambda args: args["convert"](
+                                    args["max"](
+                                        args["max"](
+                                            args["max"](args["r"], args["g"]), args["b"]
+                                        ),
+                                        args["a"],
+                                    )
+                                    * 255,
+                                    "1",
+                                ),
                                 r=r,
                                 g=g,
                                 b=b,
@@ -667,44 +699,47 @@ def _write_multiple_frames(im, fp, palette):
                                 delta_l = Image.new("L", delta.size)
                                 delta_l.putdata(delta.getdata())
                                 delta = delta_l
-                            mask = ImageMath.eval("convert(im * 255, '1')", im=delta)
+                            mask = ImageMath.lambda_eval(
+                                lambda args: args["convert"](args["im"] * 255, "1"),
+                                im=delta,
+                            )
                         diff_frame.paste(fill, mask=ImageOps.invert(mask))
             else:
                 bbox = None
             previous_im = im_frame
-            im_frames.append(
-                {"im": diff_frame or im_frame, "bbox": bbox, "encoderinfo": encoderinfo}
-            )
+            im_frames.append(_Frame(diff_frame or im_frame, bbox, encoderinfo))
 
     if len(im_frames) == 1:
         if "duration" in im.encoderinfo:
             # Since multiple frames will not be written, use the combined duration
-            im.encoderinfo["duration"] = im_frames[0]["encoderinfo"]["duration"]
-        return
+            im.encoderinfo["duration"] = im_frames[0].encoderinfo["duration"]
+        return False
 
     for frame_data in im_frames:
-        im_frame = frame_data["im"]
-        if not frame_data["bbox"]:
+        im_frame = frame_data.im
+        if not frame_data.bbox:
             # global header
-            for s in _get_global_header(im_frame, frame_data["encoderinfo"]):
+            for s in _get_global_header(im_frame, frame_data.encoderinfo):
                 fp.write(s)
             offset = (0, 0)
         else:
             # compress difference
             if not palette:
-                frame_data["encoderinfo"]["include_color_table"] = True
+                frame_data.encoderinfo["include_color_table"] = True
 
-            im_frame = im_frame.crop(frame_data["bbox"])
-            offset = frame_data["bbox"][:2]
-        _write_frame_data(fp, im_frame, offset, frame_data["encoderinfo"])
+            im_frame = im_frame.crop(frame_data.bbox)
+            offset = frame_data.bbox[:2]
+        _write_frame_data(fp, im_frame, offset, frame_data.encoderinfo)
     return True
 
 
-def _save_all(im, fp, filename):
+def _save_all(im: Image.Image, fp: IO[bytes], filename: str | bytes) -> None:
     _save(im, fp, filename, save_all=True)
 
 
-def _save(im, fp, filename, save_all=False):
+def _save(
+    im: Image.Image, fp: IO[bytes], filename: str | bytes, save_all: bool = False
+) -> None:
     # header
     if "palette" in im.encoderinfo or "palette" in im.info:
         palette = im.encoderinfo.get("palette", im.info.get("palette"))
@@ -721,7 +756,7 @@ def _save(im, fp, filename, save_all=False):
         fp.flush()
 
 
-def get_interlace(im):
+def get_interlace(im: Image.Image) -> int:
     interlace = im.encoderinfo.get("interlace", 1)
 
     # workaround for @PIL153
@@ -731,7 +766,9 @@ def get_interlace(im):
     return interlace
 
 
-def _write_local_header(fp, im, offset, flags):
+def _write_local_header(
+    fp: IO[bytes], im: Image.Image, offset: tuple[int, int], flags: int
+) -> None:
     try:
         transparency = im.encoderinfo["transparency"]
     except KeyError:
@@ -779,7 +816,7 @@ def _write_local_header(fp, im, offset, flags):
     fp.write(o8(8))  # bits
 
 
-def _save_netpbm(im, fp, filename):
+def _save_netpbm(im: Image.Image, fp: IO[bytes], filename: str | bytes) -> None:
     # Unused by default.
     # To use, uncomment the register_save call at the end of the file.
     #
@@ -810,6 +847,7 @@ def _save_netpbm(im, fp, filename):
                 )
 
                 # Allow ppmquant to receive SIGPIPE if ppmtogif exits
+                assert quant_proc.stdout is not None
                 quant_proc.stdout.close()
 
                 retcode = quant_proc.wait()
@@ -831,7 +869,7 @@ def _save_netpbm(im, fp, filename):
 _FORCE_OPTIMIZE = False
 
 
-def _get_optimize(im, info):
+def _get_optimize(im: Image.Image, info: dict[str, Any]) -> list[int] | None:
     """
     Palette optimization is a potentially expensive operation.
 
@@ -875,9 +913,10 @@ def _get_optimize(im, info):
                 and current_palette_size > 2
             ):
                 return used_palette_colors
+    return None
 
 
-def _get_color_table_size(palette_bytes):
+def _get_color_table_size(palette_bytes: bytes) -> int:
     # calculate the palette size for the header
     if not palette_bytes:
         return 0
@@ -887,7 +926,7 @@ def _get_color_table_size(palette_bytes):
         return math.ceil(math.log(len(palette_bytes) // 3, 2)) - 1
 
 
-def _get_header_palette(palette_bytes):
+def _get_header_palette(palette_bytes: bytes) -> bytes:
     """
     Returns the palette, null padded to the next power of 2 (*3) bytes
     suitable for direct inclusion in the GIF header
@@ -905,7 +944,7 @@ def _get_header_palette(palette_bytes):
     return palette_bytes
 
 
-def _get_palette_bytes(im):
+def _get_palette_bytes(im: Image.Image) -> bytes:
     """
     Gets the palette for inclusion in the gif header
 
@@ -915,7 +954,10 @@ def _get_palette_bytes(im):
     return im.palette.palette if im.palette else b""
 
 
-def _get_background(im, info_background):
+def _get_background(
+    im: Image.Image,
+    info_background: int | tuple[int, int, int] | tuple[int, int, int, int] | None,
+) -> int:
     background = 0
     if info_background:
         if isinstance(info_background, tuple):
@@ -938,7 +980,7 @@ def _get_background(im, info_background):
     return background
 
 
-def _get_global_header(im, info):
+def _get_global_header(im: Image.Image, info: dict[str, Any]) -> list[bytes]:
     """Return a list of strings representing a GIF header"""
 
     # Header Block
@@ -1000,7 +1042,12 @@ def _get_global_header(im, info):
     return header
 
 
-def _write_frame_data(fp, im_frame, offset, params):
+def _write_frame_data(
+    fp: IO[bytes],
+    im_frame: Image.Image,
+    offset: tuple[int, int],
+    params: dict[str, Any],
+) -> None:
     try:
         im_frame.encoderinfo = params
 
@@ -1020,7 +1067,9 @@ def _write_frame_data(fp, im_frame, offset, params):
 # Legacy GIF utilities
 
 
-def getheader(im, palette=None, info=None):
+def getheader(
+    im: Image.Image, palette: _Palette | None = None, info: dict[str, Any] | None = None
+) -> tuple[list[bytes], list[int] | None]:
     """
     Legacy Method to get Gif data from image.
 
@@ -1032,11 +1081,11 @@ def getheader(im, palette=None, info=None):
     :returns: tuple of(list of header items, optimized palette)
 
     """
-    used_palette_colors = _get_optimize(im, info)
-
     if info is None:
         info = {}
 
+    used_palette_colors = _get_optimize(im, info)
+
     if "background" not in info and "background" in im.info:
         info["background"] = im.info["background"]
 
@@ -1048,7 +1097,9 @@ def getheader(im, palette=None, info=None):
     return header, used_palette_colors
 
 
-def getdata(im, offset=(0, 0), **params):
+def getdata(
+    im: Image.Image, offset: tuple[int, int] = (0, 0), **params: Any
+) -> list[bytes]:
     """
     Legacy Method
 
@@ -1065,12 +1116,23 @@ def getdata(im, offset=(0, 0), **params):
     :returns: List of bytes containing GIF encoded frame data
 
     """
+    from io import BytesIO
 
-    class Collector:
+    class Collector(BytesIO):
         data = []
 
-        def write(self, data):
-            self.data.append(data)
+        if sys.version_info >= (3, 12):
+            from collections.abc import Buffer
+
+            def write(self, data: Buffer) -> int:
+                self.data.append(data)
+                return len(data)
+
+        else:
+
+            def write(self, data: Any) -> int:
+                self.data.append(data)
+                return len(data)
 
     im.load()  # make sure raster data is available
 
