@@ -16,17 +16,19 @@
 #
 # See the README file for information on usage and redistribution.
 #
+from __future__ import annotations
 
 import io
 import os
 import struct
 import sys
+from typing import IO
 
-from PIL import Image, ImageFile, PngImagePlugin, features
+from . import Image, ImageFile, PngImagePlugin, features
 
 enable_jpeg2k = features.check_codec("jpg_2000")
 if enable_jpeg2k:
-    from PIL import Jpeg2KImagePlugin
+    from . import Jpeg2KImagePlugin
 
 MAGIC = b"icns"
 HEADERSIZE = 8
@@ -42,7 +44,8 @@ def read_32t(fobj, start_length, size):
     fobj.seek(start)
     sig = fobj.read(4)
     if sig != b"\x00\x00\x00\x00":
-        raise SyntaxError("Unknown signature, expecting 0x00000000")
+        msg = "Unknown signature, expecting 0x00000000"
+        raise SyntaxError(msg)
     return read_32(fobj, (start + 4, length - 4), size)
 
 
@@ -82,7 +85,8 @@ def read_32(fobj, start_length, size):
                 if bytesleft <= 0:
                     break
             if bytesleft != 0:
-                raise SyntaxError(f"Error reading channel [{repr(bytesleft)} left]")
+                msg = f"Error reading channel [{repr(bytesleft)} left]"
+                raise SyntaxError(msg)
             band = Image.frombuffer("L", pixel_size, b"".join(data), "raw", "L", 0, 1)
             im.im.putband(band.im, band_ix)
     return {"RGB": im}
@@ -113,10 +117,11 @@ def read_png_or_jpeg2000(fobj, start_length, size):
         or sig == b"\x00\x00\x00\x0cjP  \x0d\x0a\x87\x0a"
     ):
         if not enable_jpeg2k:
-            raise ValueError(
+            msg = (
                 "Unsupported icon subimage format (rebuild PIL "
                 "with JPEG 2000 support to fix this)"
             )
+            raise ValueError(msg)
         # j2k, jpc or j2c
         fobj.seek(start)
         jp2kstream = fobj.read(length)
@@ -127,11 +132,11 @@ def read_png_or_jpeg2000(fobj, start_length, size):
             im = im.convert("RGBA")
         return {"RGBA": im}
     else:
-        raise ValueError("Unsupported icon subimage format")
+        msg = "Unsupported icon subimage format"
+        raise ValueError(msg)
 
 
 class IcnsFile:
-
     SIZES = {
         (512, 512, 2): [(b"ic10", read_png_or_jpeg2000)],
         (512, 512, 1): [(b"ic09", read_png_or_jpeg2000)],
@@ -168,12 +173,14 @@ class IcnsFile:
         self.fobj = fobj
         sig, filesize = nextheader(fobj)
         if not _accept(sig):
-            raise SyntaxError("not an icns file")
+            msg = "not an icns file"
+            raise SyntaxError(msg)
         i = HEADERSIZE
         while i < filesize:
             sig, blocksize = nextheader(fobj)
             if blocksize <= 0:
-                raise SyntaxError("invalid block header")
+                msg = "invalid block header"
+                raise SyntaxError(msg)
             i += HEADERSIZE
             blocksize -= HEADERSIZE
             dct[sig] = (i, blocksize)
@@ -183,7 +190,7 @@ class IcnsFile:
     def itersizes(self):
         sizes = []
         for size, fmts in self.SIZES.items():
-            for (fmt, reader) in fmts:
+            for fmt, reader in fmts:
                 if fmt in self.dct:
                     sizes.append(size)
                     break
@@ -192,7 +199,8 @@ class IcnsFile:
     def bestsize(self):
         sizes = self.itersizes()
         if not sizes:
-            raise SyntaxError("No 32bit icon resources found")
+            msg = "No 32bit icon resources found"
+            raise SyntaxError(msg)
         return max(sizes)
 
     def dataforsize(self, size):
@@ -245,9 +253,9 @@ class IcnsImageFile(ImageFile.ImageFile):
     format = "ICNS"
     format_description = "Mac OS icns resource"
 
-    def _open(self):
+    def _open(self) -> None:
         self.icns = IcnsFile(self.fp)
-        self.mode = "RGBA"
+        self._mode = "RGBA"
         self.info["sizes"] = self.icns.itersizes()
         self.best_size = self.icns.bestsize()
         self.size = (
@@ -275,7 +283,8 @@ class IcnsImageFile(ImageFile.ImageFile):
             if value in simple_sizes:
                 info_size = self.info["sizes"][simple_sizes.index(value)]
         if info_size not in self.info["sizes"]:
-            raise ValueError("This is not one of the allowed sizes of this image")
+            msg = "This is not one of the allowed sizes of this image"
+            raise ValueError(msg)
         self._size = value
 
     def load(self):
@@ -298,13 +307,13 @@ class IcnsImageFile(ImageFile.ImageFile):
         px = im.load()
 
         self.im = im.im
-        self.mode = im.mode
+        self._mode = im.mode
         self.size = im.size
 
         return px
 
 
-def _save(im, fp, filename):
+def _save(im: Image.Image, fp: IO[bytes], filename: str | bytes) -> None:
     """
     Saves the image as a series of PNG files,
     that are then combined into a .icns file.
@@ -338,35 +347,33 @@ def _save(im, fp, filename):
     entries = []
     for type, size in sizes.items():
         stream = size_streams[size]
-        entries.append(
-            {"type": type, "size": HEADERSIZE + len(stream), "stream": stream}
-        )
+        entries.append((type, HEADERSIZE + len(stream), stream))
 
     # Header
     fp.write(MAGIC)
     file_length = HEADERSIZE  # Header
     file_length += HEADERSIZE + 8 * len(entries)  # TOC
-    file_length += sum(entry["size"] for entry in entries)
+    file_length += sum(entry[1] for entry in entries)
     fp.write(struct.pack(">i", file_length))
 
     # TOC
     fp.write(b"TOC ")
     fp.write(struct.pack(">i", HEADERSIZE + len(entries) * HEADERSIZE))
     for entry in entries:
-        fp.write(entry["type"])
-        fp.write(struct.pack(">i", entry["size"]))
+        fp.write(entry[0])
+        fp.write(struct.pack(">i", entry[1]))
 
     # Data
     for entry in entries:
-        fp.write(entry["type"])
-        fp.write(struct.pack(">i", entry["size"]))
-        fp.write(entry["stream"])
+        fp.write(entry[0])
+        fp.write(struct.pack(">i", entry[1]))
+        fp.write(entry[2])
 
     if hasattr(fp, "flush"):
         fp.flush()
 
 
-def _accept(prefix):
+def _accept(prefix: bytes) -> bool:
     return prefix[:4] == MAGIC
 
 
@@ -384,8 +391,8 @@ if __name__ == "__main__":
     with open(sys.argv[1], "rb") as fp:
         imf = IcnsImageFile(fp)
         for size in imf.info["sizes"]:
-            imf.size = size
-            imf.save("out-%s-%s-%s.png" % size)
+            width, height, scale = imf.size = size
+            imf.save(f"out-{width}-{height}-{scale}.png")
         with Image.open(sys.argv[1]) as im:
             im.save("out.png")
         if sys.platform == "windows":
