@@ -15,6 +15,9 @@
 #
 from __future__ import annotations
 
+import math
+from typing import IO
+
 from . import Image, ImageFile
 from ._binary import i16be as i16
 from ._binary import o8
@@ -35,6 +38,7 @@ MODES = {
     b"P6": "RGB",
     # extensions
     b"P0CMYK": "CMYK",
+    b"Pf": "F",
     # PIL extensions (for test purposes only)
     b"PyP": "P",
     b"PyRGBA": "RGBA",
@@ -42,8 +46,8 @@ MODES = {
 }
 
 
-def _accept(prefix):
-    return prefix[0:1] == b"P" and prefix[1] in b"0123456y"
+def _accept(prefix: bytes) -> bool:
+    return prefix[0:1] == b"P" and prefix[1] in b"0123456fy"
 
 
 ##
@@ -54,7 +58,9 @@ class PpmImageFile(ImageFile.ImageFile):
     format = "PPM"
     format_description = "Pbmplus image"
 
-    def _read_magic(self):
+    def _read_magic(self) -> bytes:
+        assert self.fp is not None
+
         magic = b""
         # read until whitespace or longest available magic number
         for _ in range(6):
@@ -64,7 +70,9 @@ class PpmImageFile(ImageFile.ImageFile):
             magic += c
         return magic
 
-    def _read_token(self):
+    def _read_token(self) -> bytes:
+        assert self.fp is not None
+
         token = b""
         while len(token) <= 10:  # read until next whitespace or limit of 10 characters
             c = self.fp.read(1)
@@ -90,13 +98,16 @@ class PpmImageFile(ImageFile.ImageFile):
             raise ValueError(msg)
         return token
 
-    def _open(self):
+    def _open(self) -> None:
+        assert self.fp is not None
+
         magic_number = self._read_magic()
         try:
             mode = MODES[magic_number]
         except KeyError:
             msg = "not a PPM file"
             raise SyntaxError(msg)
+        self._mode = mode
 
         if magic_number in (b"P1", b"P4"):
             self.custom_mimetype = "image/x-portable-bitmap"
@@ -105,40 +116,42 @@ class PpmImageFile(ImageFile.ImageFile):
         elif magic_number in (b"P3", b"P6"):
             self.custom_mimetype = "image/x-portable-pixmap"
 
-        maxval = None
+        self._size = int(self._read_token()), int(self._read_token())
+
         decoder_name = "raw"
         if magic_number in (b"P1", b"P2", b"P3"):
             decoder_name = "ppm_plain"
-        for ix in range(3):
-            token = int(self._read_token())
-            if ix == 0:  # token is the x size
-                xsize = token
-            elif ix == 1:  # token is the y size
-                ysize = token
-                if mode == "1":
-                    self._mode = "1"
-                    rawmode = "1;I"
-                    break
-                else:
-                    self._mode = rawmode = mode
-            elif ix == 2:  # token is maxval
-                maxval = token
-                if not 0 < maxval < 65536:
-                    msg = "maxval must be greater than 0 and less than 65536"
-                    raise ValueError(msg)
-                if maxval > 255 and mode == "L":
-                    self._mode = "I"
 
-                if decoder_name != "ppm_plain":
-                    # If maxval matches a bit depth, use the raw decoder directly
-                    if maxval == 65535 and mode == "L":
-                        rawmode = "I;16B"
-                    elif maxval != 255:
-                        decoder_name = "ppm"
+        args: str | tuple[str | int, ...]
+        if mode == "1":
+            args = "1;I"
+        elif mode == "F":
+            scale = float(self._read_token())
+            if scale == 0.0 or not math.isfinite(scale):
+                msg = "scale must be finite and non-zero"
+                raise ValueError(msg)
+            self.info["scale"] = abs(scale)
 
-        args = (rawmode, 0, 1) if decoder_name == "raw" else (rawmode, maxval)
-        self._size = xsize, ysize
-        self.tile = [(decoder_name, (0, 0, xsize, ysize), self.fp.tell(), args)]
+            rawmode = "F;32F" if scale < 0 else "F;32BF"
+            args = (rawmode, 0, -1)
+        else:
+            maxval = int(self._read_token())
+            if not 0 < maxval < 65536:
+                msg = "maxval must be greater than 0 and less than 65536"
+                raise ValueError(msg)
+            if maxval > 255 and mode == "L":
+                self._mode = "I"
+
+            rawmode = mode
+            if decoder_name != "ppm_plain":
+                # If maxval matches a bit depth, use the raw decoder directly
+                if maxval == 65535 and mode == "L":
+                    rawmode = "I;16B"
+                elif maxval != 255:
+                    decoder_name = "ppm"
+
+            args = rawmode if decoder_name == "raw" else (rawmode, maxval)
+        self.tile = [(decoder_name, (0, 0) + self.size, self.fp.tell(), args)]
 
 
 #
@@ -147,16 +160,19 @@ class PpmImageFile(ImageFile.ImageFile):
 
 class PpmPlainDecoder(ImageFile.PyDecoder):
     _pulls_fd = True
+    _comment_spans: bool
+
+    def _read_block(self) -> bytes:
+        assert self.fd is not None
 
-    def _read_block(self):
         return self.fd.read(ImageFile.SAFEBLOCK)
 
-    def _find_comment_end(self, block, start=0):
+    def _find_comment_end(self, block: bytes, start: int = 0) -> int:
         a = block.find(b"\n", start)
         b = block.find(b"\r", start)
         return min(a, b) if a * b > 0 else max(a, b)  # lowest nonnegative index (or -1)
 
-    def _ignore_comments(self, block):
+    def _ignore_comments(self, block: bytes) -> bytes:
         if self._comment_spans:
             # Finish current comment
             while block:
@@ -190,7 +206,7 @@ class PpmPlainDecoder(ImageFile.PyDecoder):
                 break
         return block
 
-    def _decode_bitonal(self):
+    def _decode_bitonal(self) -> bytearray:
         """
         This is a separate method because in the plain PBM format, all data tokens are
         exactly one byte, so the inter-token whitespace is optional.
@@ -215,7 +231,7 @@ class PpmPlainDecoder(ImageFile.PyDecoder):
         invert = bytes.maketrans(b"01", b"\xFF\x00")
         return data.translate(invert)
 
-    def _decode_blocks(self, maxval):
+    def _decode_blocks(self, maxval: int) -> bytearray:
         data = bytearray()
         max_len = 10
         out_byte_count = 4 if self.mode == "I" else 1
@@ -223,7 +239,7 @@ class PpmPlainDecoder(ImageFile.PyDecoder):
         bands = Image.getmodebands(self.mode)
         total_bytes = self.state.xsize * self.state.ysize * bands * out_byte_count
 
-        half_token = False
+        half_token = b""
         while len(data) != total_bytes:
             block = self._read_block()  # read next block
             if not block:
@@ -237,7 +253,7 @@ class PpmPlainDecoder(ImageFile.PyDecoder):
 
             if half_token:
                 block = half_token + block  # stitch half_token to new block
-                half_token = False
+                half_token = b""
 
             tokens = block.split()
 
@@ -254,16 +270,19 @@ class PpmPlainDecoder(ImageFile.PyDecoder):
                     msg = b"Token too long found in data: %s" % token[: max_len + 1]
                     raise ValueError(msg)
                 value = int(token)
+                if value < 0:
+                    msg_str = f"Channel value is negative: {value}"
+                    raise ValueError(msg_str)
                 if value > maxval:
-                    msg = f"Channel value too large for this mode: {value}"
-                    raise ValueError(msg)
+                    msg_str = f"Channel value too large for this mode: {value}"
+                    raise ValueError(msg_str)
                 value = round(value / maxval * out_max)
                 data += o32(value) if self.mode == "I" else o8(value)
                 if len(data) == total_bytes:  # finished!
                     break
         return data
 
-    def decode(self, buffer):
+    def decode(self, buffer: bytes) -> tuple[int, int]:
         self._comment_spans = False
         if self.mode == "1":
             data = self._decode_bitonal()
@@ -279,14 +298,17 @@ class PpmPlainDecoder(ImageFile.PyDecoder):
 class PpmDecoder(ImageFile.PyDecoder):
     _pulls_fd = True
 
-    def decode(self, buffer):
+    def decode(self, buffer: bytes) -> tuple[int, int]:
+        assert self.fd is not None
+
         data = bytearray()
         maxval = self.args[-1]
         in_byte_count = 1 if maxval < 256 else 2
         out_byte_count = 4 if self.mode == "I" else 1
         out_max = 65535 if self.mode == "I" else 255
         bands = Image.getmodebands(self.mode)
-        while len(data) < self.state.xsize * self.state.ysize * bands * out_byte_count:
+        dest_length = self.state.xsize * self.state.ysize * bands * out_byte_count
+        while len(data) < dest_length:
             pixels = self.fd.read(in_byte_count * bands)
             if len(pixels) < in_byte_count * bands:
                 # eof
@@ -306,7 +328,7 @@ class PpmDecoder(ImageFile.PyDecoder):
 # --------------------------------------------------------------------
 
 
-def _save(im, fp, filename):
+def _save(im: Image.Image, fp: IO[bytes], filename: str | bytes) -> None:
     if im.mode == "1":
         rawmode, head = "1;I", b"P4"
     elif im.mode == "L":
@@ -315,6 +337,8 @@ def _save(im, fp, filename):
         rawmode, head = "I;16B", b"P5"
     elif im.mode in ("RGB", "RGBA"):
         rawmode, head = "RGB", b"P6"
+    elif im.mode == "F":
+        rawmode, head = "F;32F", b"Pf"
     else:
         msg = f"cannot write mode {im.mode} as PPM"
         raise OSError(msg)
@@ -326,7 +350,10 @@ def _save(im, fp, filename):
             fp.write(b"255\n")
         else:
             fp.write(b"65535\n")
-    ImageFile._save(im, fp, [("raw", (0, 0) + im.size, 0, (rawmode, 0, 1))])
+    elif head == b"Pf":
+        fp.write(b"-1.0\n")
+    row_order = -1 if im.mode == "F" else 1
+    ImageFile._save(im, fp, [("raw", (0, 0) + im.size, 0, (rawmode, 0, row_order))])
 
 
 #
@@ -339,6 +366,6 @@ Image.register_save(PpmImageFile.format, _save)
 Image.register_decoder("ppm", PpmDecoder)
 Image.register_decoder("ppm_plain", PpmPlainDecoder)
 
-Image.register_extensions(PpmImageFile.format, [".pbm", ".pgm", ".ppm", ".pnm"])
+Image.register_extensions(PpmImageFile.format, [".pbm", ".pgm", ".ppm", ".pnm", ".pfm"])
 
 Image.register_mime(PpmImageFile.format, "image/x-portable-anymap")
