@@ -13,16 +13,20 @@
 
 # Notes:
 #
-#  * Implements the pixel access object following Access.
-#  * Does not implement the line functions, as they don't appear to be used
+#  * Implements the pixel access object following Access.c
 #  * Taking only the tuple form, which is used from python.
 #    * Fill.c uses the integer form, but it's still going to use the old
 #      Access.c implementation.
 #
+from __future__ import annotations
 
 import logging
 import sys
+from typing import TYPE_CHECKING
 
+from ._deprecate import deprecate
+
+FFI: type
 try:
     from cffi import FFI
 
@@ -39,15 +43,19 @@ try:
 except ImportError as ex:
     # Allow error import for doc purposes, but error out when accessing
     # anything in core.
-    from ._util import deferred_error
+    from ._util import DeferredError
 
-    FFI = ffi = deferred_error(ex)
+    FFI = ffi = DeferredError.new(ex)
 
 logger = logging.getLogger(__name__)
 
+if TYPE_CHECKING:
+    from . import Image
+
 
 class PyAccess:
-    def __init__(self, img, readonly=False):
+    def __init__(self, img: Image.Image, readonly: bool = False) -> None:
+        deprecate("PyAccess", 11)
         vals = dict(img.im.unsafe_ptrs)
         self.readonly = readonly
         self.image8 = ffi.cast("unsigned char **", vals["image8"])
@@ -58,7 +66,7 @@ class PyAccess:
 
         # Keep pointer to im object to prevent dereferencing.
         self._im = img.im
-        if self._im.mode == "P":
+        if self._im.mode in ("P", "PA"):
             self._palette = img.palette
 
         # Debugging is polluting test traces, only useful here
@@ -66,21 +74,27 @@ class PyAccess:
         # logger.debug("%s", vals)
         self._post_init()
 
-    def _post_init(self):
+    def _post_init(self) -> None:
         pass
 
-    def __setitem__(self, xy, color):
+    def __setitem__(
+        self,
+        xy: tuple[int, int] | list[int],
+        color: float | tuple[int, ...] | list[int],
+    ) -> None:
         """
         Modifies the pixel at x,y. The color is given as a single
         numerical value for single band images, and a tuple for
-        multi-band images
+        multi-band images. In addition to this, RGB and RGBA tuples
+        are accepted for P and PA images.
 
         :param xy: The pixel coordinate, given as (x, y). See
            :ref:`coordinate-system`.
         :param color: The pixel value.
         """
         if self.readonly:
-            raise ValueError("Attempt to putpixel a read only image")
+            msg = "Attempt to putpixel a read only image"
+            raise ValueError(msg)
         (x, y) = xy
         if x < 0:
             x = self.xsize + x
@@ -89,16 +103,20 @@ class PyAccess:
         (x, y) = self.check_xy((x, y))
 
         if (
-            self._im.mode == "P"
+            self._im.mode in ("P", "PA")
             and isinstance(color, (list, tuple))
             and len(color) in [3, 4]
         ):
-            # RGB or RGBA value for a P image
-            color = self._palette.getcolor(color, self._img)
+            # RGB or RGBA value for a P or PA image
+            if self._im.mode == "PA":
+                alpha = color[3] if len(color) == 4 else 255
+                color = color[:3]
+            palette_index = self._palette.getcolor(color, self._img)
+            color = (palette_index, alpha) if self._im.mode == "PA" else palette_index
 
         return self.set_pixel(x, y, color)
 
-    def __getitem__(self, xy):
+    def __getitem__(self, xy: tuple[int, int] | list[int]) -> float | tuple[int, ...]:
         """
         Returns the pixel at x,y. The pixel is returned as a single
         value for single band images or a tuple for multiple band
@@ -120,22 +138,31 @@ class PyAccess:
     putpixel = __setitem__
     getpixel = __getitem__
 
-    def check_xy(self, xy):
+    def check_xy(self, xy: tuple[int, int]) -> tuple[int, int]:
         (x, y) = xy
         if not (0 <= x < self.xsize and 0 <= y < self.ysize):
-            raise ValueError("pixel location out of range")
+            msg = "pixel location out of range"
+            raise ValueError(msg)
         return xy
 
+    def get_pixel(self, x: int, y: int) -> float | tuple[int, ...]:
+        raise NotImplementedError()
+
+    def set_pixel(
+        self, x: int, y: int, color: float | tuple[int, ...] | list[int]
+    ) -> None:
+        raise NotImplementedError()
+
 
 class _PyAccess32_2(PyAccess):
-    """ PA, LA, stored in first and last bytes of a 32 bit word """
+    """PA, LA, stored in first and last bytes of a 32 bit word"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = ffi.cast("struct Pixel_RGBA **", self.image32)
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> tuple[int, int]:
         pixel = self.pixels[y][x]
-        return (pixel.r, pixel.a)
+        return pixel.r, pixel.a
 
     def set_pixel(self, x, y, color):
         pixel = self.pixels[y][x]
@@ -145,14 +172,14 @@ class _PyAccess32_2(PyAccess):
 
 
 class _PyAccess32_3(PyAccess):
-    """ RGB and friends, stored in the first three bytes of a 32 bit word """
+    """RGB and friends, stored in the first three bytes of a 32 bit word"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = ffi.cast("struct Pixel_RGBA **", self.image32)
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> tuple[int, int, int]:
         pixel = self.pixels[y][x]
-        return (pixel.r, pixel.g, pixel.b)
+        return pixel.r, pixel.g, pixel.b
 
     def set_pixel(self, x, y, color):
         pixel = self.pixels[y][x]
@@ -164,14 +191,14 @@ class _PyAccess32_3(PyAccess):
 
 
 class _PyAccess32_4(PyAccess):
-    """ RGBA etc, all 4 bytes of a 32 bit word """
+    """RGBA etc, all 4 bytes of a 32 bit word"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = ffi.cast("struct Pixel_RGBA **", self.image32)
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> tuple[int, int, int, int]:
         pixel = self.pixels[y][x]
-        return (pixel.r, pixel.g, pixel.b, pixel.a)
+        return pixel.r, pixel.g, pixel.b, pixel.a
 
     def set_pixel(self, x, y, color):
         pixel = self.pixels[y][x]
@@ -183,12 +210,12 @@ class _PyAccess32_4(PyAccess):
 
 
 class _PyAccess8(PyAccess):
-    """ 1, L, P, 8 bit images stored as uint8 """
+    """1, L, P, 8 bit images stored as uint8"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = self.image8
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> int:
         return self.pixels[y][x]
 
     def set_pixel(self, x, y, color):
@@ -201,12 +228,12 @@ class _PyAccess8(PyAccess):
 
 
 class _PyAccessI16_N(PyAccess):
-    """ I;16 access, native bitendian without conversion """
+    """I;16 access, native bitendian without conversion"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = ffi.cast("unsigned short **", self.image)
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> int:
         return self.pixels[y][x]
 
     def set_pixel(self, x, y, color):
@@ -219,12 +246,12 @@ class _PyAccessI16_N(PyAccess):
 
 
 class _PyAccessI16_L(PyAccess):
-    """ I;16L access, with conversion """
+    """I;16L access, with conversion"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = ffi.cast("struct Pixel_I16 **", self.image)
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> int:
         pixel = self.pixels[y][x]
         return pixel.l + pixel.r * 256
 
@@ -235,17 +262,17 @@ class _PyAccessI16_L(PyAccess):
         except TypeError:
             color = min(color[0], 65535)
 
-        pixel.l = color & 0xFF  # noqa: E741
+        pixel.l = color & 0xFF
         pixel.r = color >> 8
 
 
 class _PyAccessI16_B(PyAccess):
-    """ I;16B access, with conversion """
+    """I;16B access, with conversion"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = ffi.cast("struct Pixel_I16 **", self.image)
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> int:
         pixel = self.pixels[y][x]
         return pixel.l * 256 + pixel.r
 
@@ -256,17 +283,17 @@ class _PyAccessI16_B(PyAccess):
         except Exception:
             color = min(color[0], 65535)
 
-        pixel.l = color >> 8  # noqa: E741
+        pixel.l = color >> 8
         pixel.r = color & 0xFF
 
 
 class _PyAccessI32_N(PyAccess):
-    """ Signed Int32 access, native endian """
+    """Signed Int32 access, native endian"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = self.image32
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> int:
         return self.pixels[y][x]
 
     def set_pixel(self, x, y, color):
@@ -274,7 +301,7 @@ class _PyAccessI32_N(PyAccess):
 
 
 class _PyAccessI32_Swap(PyAccess):
-    """ I;32L/B access, with byteswapping conversion """
+    """I;32L/B access, with byteswapping conversion"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = self.image32
@@ -285,7 +312,7 @@ class _PyAccessI32_Swap(PyAccess):
         chars[0], chars[1], chars[2], chars[3] = chars[3], chars[2], chars[1], chars[0]
         return ffi.cast("int *", chars)[0]
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> int:
         return self.reverse(self.pixels[y][x])
 
     def set_pixel(self, x, y, color):
@@ -293,12 +320,12 @@ class _PyAccessI32_Swap(PyAccess):
 
 
 class _PyAccessF(PyAccess):
-    """ 32 bit float access """
+    """32 bit float access"""
 
     def _post_init(self, *args, **kwargs):
         self.pixels = ffi.cast("float **", self.image32)
 
-    def get_pixel(self, x, y):
+    def get_pixel(self, x: int, y: int) -> float:
         return self.pixels[y][x]
 
     def set_pixel(self, x, y, color):
@@ -314,6 +341,7 @@ mode_map = {
     "1": _PyAccess8,
     "L": _PyAccess8,
     "P": _PyAccess8,
+    "I;16N": _PyAccessI16_N,
     "LA": _PyAccess32_2,
     "La": _PyAccess32_2,
     "PA": _PyAccess32_2,
@@ -345,7 +373,7 @@ else:
     mode_map["I;32B"] = _PyAccessI32_N
 
 
-def new(img, readonly=False):
+def new(img: Image.Image, readonly: bool = False) -> PyAccess | None:
     access_type = mode_map.get(img.mode, None)
     if not access_type:
         logger.debug("PyAccess Not Implemented: %s", img.mode)
